@@ -1,5 +1,7 @@
 package Graph::TransitiveClosure;
 
+use strict;
+
 # COMMENT THESE OUT FOR TESTING AND PRODUCTION.
 # $SIG{__DIE__ } = sub { use Carp; confess };
 # $SIG{__WARN__} = sub { use Carp; confess };
@@ -61,6 +63,8 @@ sub is_transitive {
 __END__
 =pod
 
+=head1 NAME
+
 Graph::TransitiveClosure - create and query transitive closure of graph
 
 =head1 SYNOPSIS
